#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# --- BEGIN_HEADER ---
#
# pygdb.console.core - python gdb console core functions
# Copyright (C) 2019-2020  The pygdb Project lead by Brian Vinter
#
# This file is part of pygdb.
#
# pygdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pygdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# -- END_HEADER ---
#

#
# This file is a modified version of python-gdb.py from python 2.7.5:
# https://devguide.python.org/gdb/#gdb-7-and-later
#
# You can redistribute it and/or modify it under the terms of the
# PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2:
# https://docs.python.org/2.7/license.html
#
# From gdb 7 onwards, gdb's build can be configured --with-python, allowing gdb
# to be extended with Python code e.g. for library-specific data visualizations,
# such as for the C++ STL types.  Documentation on this API can be seen at:
# http://sourceware.org/gdb/current/onlinedocs/gdb/Python-API.html
#
#
# This python module deals with the case when the process being debugged (the
# "inferior process" in gdb parlance) is itself python, or more specifically,
# linked against libpython.  In this situation, almost every item of data is a
# (PyObject*), and having the debugger merely print their addresses is not very
# enlightening.
#
# This module embeds knowledge about the implementation details of libpython so
# that we can emit useful visualizations e.g. a string, a list, a dict, a frame
# giving file/line information and the state of local variables
#
# In particular, given a gdb.Value corresponding to a PyObject* in the inferior
# process, we can generate a "proxy value" within the gdb process.  For example,
# given a PyObject* in the inferior process that is in fact a PyListObject*
# holding three PyObject* that turn out to be PyStringObject* instances, we can
# generate a proxy value within the gdb process that is a list of strings:
#   ["foo", "bar", "baz"]

# Doing so can be expensive for complicated graphs of objects, and could take
# some time, so we also have a "write_repr" method that writes a representation
# of the data to a file-like object.  This allows us to stop the traversal by
# having the file-like object raise an exception if it gets too much data.
#
# With both "proxyval" and "write_repr" we keep track of the set of all addresses
# visited so far in the traversal, to avoid infinite recursion due to cycles in
# the graph of object references.
#
# We try to defer gdb.lookup_type() invocations for python types until as late as
# possible: for a dynamically linked python binary, when the process starts in
# the debugger, the libpython.so hasn't been dynamically loaded yet, so none of
# the type names are known to the debugger
#
# The module also extends gdb with some python-specific commands.
#

"""GDB console python core functions"""

import sys
import gdb

# Look up the gdb.Type for some standard types:
_type_char_ptr = gdb.lookup_type('char').pointer()  # char*
_type_unsigned_char_ptr = gdb.lookup_type(
    'unsigned char').pointer()  # unsigned char*
_type_void_ptr = gdb.lookup_type('void').pointer()  # void*
_type_size_t = gdb.lookup_type('size_t')

SIZEOF_VOID_P = _type_void_ptr.sizeof


Py_TPFLAGS_HEAPTYPE = (1L << 9)

Py_TPFLAGS_INT_SUBCLASS = (1L << 23)
Py_TPFLAGS_LONG_SUBCLASS = (1L << 24)
Py_TPFLAGS_LIST_SUBCLASS = (1L << 25)
Py_TPFLAGS_TUPLE_SUBCLASS = (1L << 26)
Py_TPFLAGS_STRING_SUBCLASS = (1L << 27)
Py_TPFLAGS_UNICODE_SUBCLASS = (1L << 28)
Py_TPFLAGS_DICT_SUBCLASS = (1L << 29)
Py_TPFLAGS_BASE_EXC_SUBCLASS = (1L << 30)
Py_TPFLAGS_TYPE_SUBCLASS = (1L << 31)


MAX_OUTPUT_LEN = 1024


class NullPyObjectPtr(RuntimeError):
    pass


def safety_limit(val):
    # Given a integer value from the process being debugged, limit it to some
    # safety threshold so that arbitrary breakage within said process doesn't
    # break the gdb process too much (e.g. sizes of iterations, sizes of lists)
    return min(val, 1000)


def safe_range(val):
    # As per range, but don't trust the value too much: cap it to a safety
    # threshold in case the data was corrupted
    return xrange(safety_limit(val))


class StringTruncated(RuntimeError):
    pass


class TruncatedStringIO(object):
    """Similar to cStringIO, but can truncate the output by raising a
    StringTruncated exception"""

    def __init__(self, maxlen=None):
        self._val = ''
        self.maxlen = maxlen

    def write(self, data):
        if self.maxlen:
            if len(data) + len(self._val) > self.maxlen:
                # Truncation:
                self._val += data[0:self.maxlen - len(self._val)]
                raise StringTruncated()

        self._val += data

    def getvalue(self):
        return self._val


class PyObjectPtr(object):
    """
    Class wrapping a gdb.Value that's a either a (PyObject*) within the
    inferior process, or some subclass pointer e.g. (PyStringObject*)

    There will be a subclass for every refined PyObject type that we care
    about.

    Note that at every stage the underlying pointer could be NULL, point
    to corrupt data, etc; this is the debugger, after all.
    """
    _typename = 'PyObject'

    def __init__(self, gdbval, cast_to=None):
        if cast_to:
            self._gdbval = gdbval.cast(cast_to)
        else:
            self._gdbval = gdbval

    def field(self, name):
        """
        Get the gdb.Value for the given field within the PyObject, coping with
        some python 2 versus python 3 differences.

        Various libpython types are defined using the "PyObject_HEAD" and
        "PyObject_VAR_HEAD" macros.

        In Python 2, this these are defined so that "ob_type" and (for a var
        object) "ob_size" are fields of the type in question.

        In Python 3, this is defined as an embedded PyVarObject type thus:
           PyVarObject ob_base;
        so that the "ob_size" field is located insize the "ob_base" field, and
        the "ob_type" is most easily accessed by casting back to a (PyObject*).
        """
        if self.is_null():
            raise NullPyObjectPtr(self)

        if name == 'ob_type':
            pyo_ptr = self._gdbval.cast(PyObjectPtr.get_gdb_type())
            return pyo_ptr.dereference()[name]

        if name == 'ob_size':
            try:
                # Python 2:
                return self._gdbval.dereference()[name]
            except RuntimeError:
                # Python 3:
                return self._gdbval.dereference()['ob_base'][name]

        # General case: look it up inside the object:
        return self._gdbval.dereference()[name]

    def pyop_field(self, name):
        """
        Get a PyObjectPtr for the given PyObject* field within this PyObject,
        coping with some python 2 versus python 3 differences.
        """
        return PyObjectPtr.from_pyobject_ptr(self.field(name))

    def write_field_repr(self, name, out, visited):
        """
        Extract the PyObject* field named "name", and write its representation
        to file-like object "out"
        """
        field_obj = self.pyop_field(name)
        field_obj.write_repr(out, visited)

    def get_truncated_repr(self, maxlen):
        """
        Get a repr-like string for the data, but truncate it at "maxlen" bytes
        (ending the object graph traversal as soon as you do)
        """
        out = TruncatedStringIO(maxlen)
        try:
            self.write_repr(out, set())
        except StringTruncated:
            # Truncation occurred:
            return out.getvalue() + '...(truncated)'

        # No truncation occurred:
        return out.getvalue()

    def type(self):
        return PyTypeObjectPtr(self.field('ob_type'))

    def is_null(self):
        return 0 == long(self._gdbval)

    def is_optimized_out(self):
        """
        Is the value of the underlying PyObject* visible to the debugger?

        This can vary with the precise version of the compiler used to build
        Python, and the precise version of gdb.

        See e.g. https://bugzilla.redhat.com/show_bug.cgi?id=556975 with
        PyEval_EvalFrameEx's "f"
        """
        return self._gdbval.is_optimized_out

    def safe_tp_name(self):
        try:
            return self.type().field('tp_name').string()
        except NullPyObjectPtr:
            # NULL tp_name?
            return 'unknown'
        except RuntimeError:
            # Can't even read the object at all?
            return 'unknown'

    def proxyval(self, visited):
        """
        Scrape a value from the inferior process, and try to represent it
        within the gdb process, whilst (hopefully) avoiding crashes when
        the remote data is corrupt.

        Derived classes will override this.

        For example, a PyIntObject* with ob_ival 42 in the inferior process
        should result in an int(42) in this process.

        visited: a set of all gdb.Value pyobject pointers already visited
        whilst generating this value (to guard against infinite recursion when
        visiting object graphs with loops).  Analogous to Py_ReprEnter and
        Py_ReprLeave
        """

        class FakeRepr(object):
            """
            Class representing a non-descript PyObject* value in the inferior
            process for when we don't have a custom scraper, intended to have
            a sane repr().
            """

            def __init__(self, tp_name, address):
                self.tp_name = tp_name
                self.address = address

            def __repr__(self):
                # For the NULL pointer, we have no way of knowing a type, so
                # special-case it as per
                # http://bugs.python.org/issue8032#msg100882
                if self.address == 0:
                    return '0x0'
                return '<%s at remote 0x%x>' % (self.tp_name, self.address)

        return FakeRepr(self.safe_tp_name(),
                        long(self._gdbval))

    def write_repr(self, out, visited):
        """
        Write a string representation of the value scraped from the inferior
        process to "out", a file-like object.
        """
        # Default implementation: generate a proxy value and write its repr
        # However, this could involve a lot of work for complicated objects,
        # so for derived classes we specialize this
        return out.write(repr(self.proxyval(visited)))

    @classmethod
    def subclass_from_type(cls, t):
        """
        Given a PyTypeObjectPtr instance wrapping a gdb.Value that's a
        (PyTypeObject*), determine the corresponding subclass of PyObjectPtr
        to use

        Ideally, we would look up the symbols for the global types, but that
        isn't working yet:
          (gdb) python print gdb.lookup_symbol('PyList_Type')[0].value
          Traceback (most recent call last):
            File "<string>", line 1, in <module>
          NotImplementedError: Symbol type not yet supported in Python scripts.
          Error while executing Python code.

        For now, we use tp_flags, after doing some string comparisons on the
        tp_name for some special-cases that don't seem to be visible through
        flags
        """
        try:
            tp_name = t.field('tp_name').string()
            tp_flags = int(t.field('tp_flags'))
        except RuntimeError:
            # Handle any kind of error e.g. NULL ptrs by simply using the base
            # class
            return cls

        # print 'tp_flags = 0x%08x' % tp_flags
        # print 'tp_name = %r' % tp_name

        name_map = {'bool': PyBoolObjectPtr,
                    'classobj': PyClassObjectPtr,
                    'instance': PyInstanceObjectPtr,
                    'NoneType': PyNoneStructPtr,
                    'frame': PyFrameObjectPtr,
                    'set': PySetObjectPtr,
                    'frozenset': PySetObjectPtr,
                    'builtin_function_or_method': PyCFunctionObjectPtr,
                    }
        if tp_name in name_map:
            return name_map[tp_name]

        if tp_flags & Py_TPFLAGS_HEAPTYPE:
            return HeapTypeObjectPtr

        if tp_flags & Py_TPFLAGS_INT_SUBCLASS:
            return PyIntObjectPtr
        if tp_flags & Py_TPFLAGS_LONG_SUBCLASS:
            return PyLongObjectPtr
        if tp_flags & Py_TPFLAGS_LIST_SUBCLASS:
            return PyListObjectPtr
        if tp_flags & Py_TPFLAGS_TUPLE_SUBCLASS:
            return PyTupleObjectPtr
        if tp_flags & Py_TPFLAGS_STRING_SUBCLASS:
            return PyStringObjectPtr
        if tp_flags & Py_TPFLAGS_UNICODE_SUBCLASS:
            return PyUnicodeObjectPtr
        if tp_flags & Py_TPFLAGS_DICT_SUBCLASS:
            return PyDictObjectPtr
        if tp_flags & Py_TPFLAGS_BASE_EXC_SUBCLASS:
            return PyBaseExceptionObjectPtr
        # if tp_flags & Py_TPFLAGS_TYPE_SUBCLASS:
        #    return PyTypeObjectPtr

        # Use the base class:
        return cls

    @classmethod
    def from_pyobject_ptr(cls, gdbval):
        """
        Try to locate the appropriate derived class dynamically, and cast
        the pointer accordingly.
        """
        try:
            p = PyObjectPtr(gdbval)
            cls = cls.subclass_from_type(p.type())
            return cls(gdbval, cast_to=cls.get_gdb_type())
        except RuntimeError:
            # Handle any kind of error e.g. NULL ptrs by simply using the base
            # class
            pass
        return cls(gdbval)

    @classmethod
    def get_gdb_type(cls):
        return gdb.lookup_type(cls._typename).pointer()

    def as_address(self):
        return long(self._gdbval)


class ProxyAlreadyVisited(object):
    """
    Placeholder proxy to use when protecting against infinite recursion due to
    loops in the object graph.

    Analogous to the values emitted by the users of Py_ReprEnter and Py_ReprLeave
    """

    def __init__(self, rep):
        self._rep = rep

    def __repr__(self):
        return self._rep


def _write_instance_repr(out, visited, name, pyop_attrdict, address):
    """Shared code for use by old-style and new-style classes:
    write a representation to file-like object 'out'"""
    out.write('<')
    out.write(name)

    # Write dictionary of instance attributes:
    if isinstance(pyop_attrdict, PyDictObjectPtr):
        out.write('(')
        first = True
        for pyop_arg, pyop_val in pyop_attrdict.iteritems():
            if not first:
                out.write(', ')
            first = False
            out.write(pyop_arg.proxyval(visited))
            out.write('=')
            pyop_val.write_repr(out, visited)
        out.write(')')
    out.write(' at remote 0x%x>' % address)


class InstanceProxy(object):

    def __init__(self, cl_name, attrdict, address):
        self.cl_name = cl_name
        self.attrdict = attrdict
        self.address = address

    def __repr__(self):
        if isinstance(self.attrdict, dict):
            kwargs = ', '.join(["%s=%r" % (arg, val)
                                for arg, val in self.attrdict.iteritems()])
            return '<%s(%s) at remote 0x%x>' % (self.cl_name,
                                                kwargs, self.address)
        else:
            return '<%s at remote 0x%x>' % (self.cl_name,
                                            self.address)


def _PyObject_VAR_SIZE(typeobj, nitems):
    return ((typeobj.field('tp_basicsize') +
             nitems * typeobj.field('tp_itemsize') +
             (SIZEOF_VOID_P - 1)
             ) & ~(SIZEOF_VOID_P - 1)
            ).cast(_type_size_t)


class HeapTypeObjectPtr(PyObjectPtr):
    _typename = 'PyObject'

    def get_attr_dict(self):
        """
        Get the PyDictObject ptr representing the attribute dictionary
        (or None if there's a problem)
        """
        try:
            typeobj = self.type()
            dictoffset = int_from_int(typeobj.field('tp_dictoffset'))
            if dictoffset != 0:
                if dictoffset < 0:
                    type_PyVarObject_ptr = gdb.lookup_type(
                        'PyVarObject').pointer()
                    tsize = int_from_int(self._gdbval.cast(
                        type_PyVarObject_ptr)['ob_size'])
                    if tsize < 0:
                        tsize = -tsize
                    size = _PyObject_VAR_SIZE(typeobj, tsize)
                    dictoffset += size
                    assert dictoffset > 0
                    assert dictoffset % SIZEOF_VOID_P == 0

                dictptr = self._gdbval.cast(_type_char_ptr) + dictoffset
                PyObjectPtrPtr = PyObjectPtr.get_gdb_type().pointer()
                dictptr = dictptr.cast(PyObjectPtrPtr)
                return PyObjectPtr.from_pyobject_ptr(dictptr.dereference())
        except RuntimeError:
            # Corrupt data somewhere; fail safe
            pass

        # Not found, or some kind of error:
        return None

    def proxyval(self, visited):
        """
        Support for new-style classes.

        Currently we just locate the dictionary using a transliteration to
        python of _PyObject_GetDictPtr, ignoring descriptors
        """
        # Guard against infinite loops:
        if self.as_address() in visited:
            return ProxyAlreadyVisited('<...>')
        visited.add(self.as_address())

        pyop_attr_dict = self.get_attr_dict()
        if pyop_attr_dict:
            attr_dict = pyop_attr_dict.proxyval(visited)
        else:
            attr_dict = {}
        tp_name = self.safe_tp_name()

        # New-style class:
        return InstanceProxy(tp_name, attr_dict, long(self._gdbval))

    def write_repr(self, out, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            out.write('<...>')
            return
        visited.add(self.as_address())

        pyop_attrdict = self.get_attr_dict()
        _write_instance_repr(out, visited,
                             self.safe_tp_name(),
                             pyop_attrdict,
                             self.as_address())


class ProxyException(Exception):
    def __init__(self, tp_name, args):
        self.tp_name = tp_name
        self.args = args

    def __repr__(self):
        return '%s%r' % (self.tp_name, self.args)


class PyBaseExceptionObjectPtr(PyObjectPtr):
    """
    Class wrapping a gdb.Value that's a PyBaseExceptionObject* i.e. an exception
    within the process being debugged.
    """
    _typename = 'PyBaseExceptionObject'

    def proxyval(self, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            return ProxyAlreadyVisited('(...)')
        visited.add(self.as_address())
        arg_proxy = self.pyop_field('args').proxyval(visited)
        return ProxyException(self.safe_tp_name(),
                              arg_proxy)

    def write_repr(self, out, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            out.write('(...)')
            return
        visited.add(self.as_address())

        out.write(self.safe_tp_name())
        self.write_field_repr('args', out, visited)


class PyBoolObjectPtr(PyObjectPtr):
    """
    Class wrapping a gdb.Value that's a PyBoolObject* i.e. one of the two
    <bool> instances (Py_True/Py_False) within the process being debugged.
    """
    _typename = 'PyBoolObject'

    def proxyval(self, visited):
        if int_from_int(self.field('ob_ival')):
            return True
        else:
            return False


class PyClassObjectPtr(PyObjectPtr):
    """
    Class wrapping a gdb.Value that's a PyClassObject* i.e. a <classobj>
    instance within the process being debugged.
    """
    _typename = 'PyClassObject'


class BuiltInFunctionProxy(object):
    def __init__(self, ml_name):
        self.ml_name = ml_name

    def __repr__(self):
        return "<built-in function %s>" % self.ml_name


class BuiltInMethodProxy(object):
    def __init__(self, ml_name, pyop_m_self):
        self.ml_name = ml_name
        self.pyop_m_self = pyop_m_self

    def __repr__(self):
        return ('<built-in method %s of %s object at remote 0x%x>'
                % (self.ml_name,
                   self.pyop_m_self.safe_tp_name(),
                   self.pyop_m_self.as_address())
                )


class PyCFunctionObjectPtr(PyObjectPtr):
    """
    Class wrapping a gdb.Value that's a PyCFunctionObject*
    (see Include/methodobject.h and Objects/methodobject.c)
    """
    _typename = 'PyCFunctionObject'

    def proxyval(self, visited):
        m_ml = self.field('m_ml')  # m_ml is a (PyMethodDef*)
        ml_name = m_ml['ml_name'].string()

        pyop_m_self = self.pyop_field('m_self')
        if pyop_m_self.is_null():
            return BuiltInFunctionProxy(ml_name)
        else:
            return BuiltInMethodProxy(ml_name, pyop_m_self)


class PyCodeObjectPtr(PyObjectPtr):
    """
    Class wrapping a gdb.Value that's a PyCodeObject* i.e. a <code> instance
    within the process being debugged.
    """
    _typename = 'PyCodeObject'

    def addr2line(self, addrq):
        """
        Get the line number for a given bytecode offset

        Analogous to PyCode_Addr2Line; translated from pseudocode in
        Objects/lnotab_notes.txt
        """
        co_lnotab = self.pyop_field('co_lnotab').proxyval(set())

        # Initialize lineno to co_firstlineno as per PyCode_Addr2Line
        # not 0, as lnotab_notes.txt has it:
        lineno = int_from_int(self.field('co_firstlineno'))

        addr = 0
        for addr_incr, line_incr in zip(co_lnotab[::2], co_lnotab[1::2]):
            addr += ord(addr_incr)
            if addr > addrq:
                return lineno
            lineno += ord(line_incr)
        return lineno


class PyDictObjectPtr(PyObjectPtr):
    """
    Class wrapping a gdb.Value that's a PyDictObject* i.e. a dict instance
    within the process being debugged.
    """
    _typename = 'PyDictObject'

    def iteritems(self):
        """
        Yields a sequence of (PyObjectPtr key, PyObjectPtr value) pairs,
        analagous to dict.iteritems()
        """
        for i in safe_range(self.field('ma_mask') + 1):
            ep = self.field('ma_table') + i
            pyop_value = PyObjectPtr.from_pyobject_ptr(ep['me_value'])
            if not pyop_value.is_null():
                pyop_key = PyObjectPtr.from_pyobject_ptr(ep['me_key'])
                yield (pyop_key, pyop_value)

    def proxyval(self, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            return ProxyAlreadyVisited('{...}')
        visited.add(self.as_address())

        result = {}
        for pyop_key, pyop_value in self.iteritems():
            proxy_key = pyop_key.proxyval(visited)
            proxy_value = pyop_value.proxyval(visited)
            result[proxy_key] = proxy_value
        return result

    def write_repr(self, out, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            out.write('{...}')
            return
        visited.add(self.as_address())

        out.write('{')
        first = True
        for pyop_key, pyop_value in self.iteritems():
            if not first:
                out.write(', ')
            first = False
            pyop_key.write_repr(out, visited)
            out.write(': ')
            pyop_value.write_repr(out, visited)
        out.write('}')


class PyInstanceObjectPtr(PyObjectPtr):
    _typename = 'PyInstanceObject'

    def proxyval(self, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            return ProxyAlreadyVisited('<...>')
        visited.add(self.as_address())

        # Get name of class:
        in_class = self.pyop_field('in_class')
        cl_name = in_class.pyop_field('cl_name').proxyval(visited)

        # Get dictionary of instance attributes:
        in_dict = self.pyop_field('in_dict').proxyval(visited)

        # Old-style class:
        return InstanceProxy(cl_name, in_dict, long(self._gdbval))

    def write_repr(self, out, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            out.write('<...>')
            return
        visited.add(self.as_address())

        # Old-style class:

        # Get name of class:
        in_class = self.pyop_field('in_class')
        cl_name = in_class.pyop_field('cl_name').proxyval(visited)

        # Get dictionary of instance attributes:
        pyop_in_dict = self.pyop_field('in_dict')

        _write_instance_repr(out, visited,
                             cl_name, pyop_in_dict, self.as_address())


class PyIntObjectPtr(PyObjectPtr):
    _typename = 'PyIntObject'

    def proxyval(self, visited):
        result = int_from_int(self.field('ob_ival'))
        return result


class PyListObjectPtr(PyObjectPtr):
    _typename = 'PyListObject'

    def __getitem__(self, i):
        # Get the gdb.Value for the (PyObject*) with the given index:
        field_ob_item = self.field('ob_item')
        return field_ob_item[i]

    def proxyval(self, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            return ProxyAlreadyVisited('[...]')
        visited.add(self.as_address())

        result = [PyObjectPtr.from_pyobject_ptr(self[i]).proxyval(visited)
                  for i in safe_range(int_from_int(self.field('ob_size')))]
        return result

    def write_repr(self, out, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            out.write('[...]')
            return
        visited.add(self.as_address())

        out.write('[')
        for i in safe_range(int_from_int(self.field('ob_size'))):
            if i > 0:
                out.write(', ')
            element = PyObjectPtr.from_pyobject_ptr(self[i])
            element.write_repr(out, visited)
        out.write(']')


class PyLongObjectPtr(PyObjectPtr):
    _typename = 'PyLongObject'

    def proxyval(self, visited):
        """
        Python's Include/longobjrep.h has this declaration:
           struct _longobject {
               PyObject_VAR_HEAD
               digit ob_digit[1];
           };

        with this description:
            The absolute value of a number is equal to
                 SUM(for i=0 through abs(ob_size)-1) ob_digit[i] * 2**(SHIFT*i)
            Negative numbers are represented with ob_size < 0;
            zero is represented by ob_size == 0.

        where SHIFT can be either:
            #define PyLong_SHIFT        30
            #define PyLong_SHIFT        15
        """
        ob_size = long(self.field('ob_size'))
        if ob_size == 0:
            return 0L

        ob_digit = self.field('ob_digit')

        if gdb.lookup_type('digit').sizeof == 2:
            SHIFT = 15L
        else:
            SHIFT = 30L

        digits = [long(ob_digit[i]) * 2**(SHIFT*i)
                  for i in safe_range(abs(ob_size))]
        result = sum(digits)
        if ob_size < 0:
            result = -result
        return result


class PyNoneStructPtr(PyObjectPtr):
    """
    Class wrapping a gdb.Value that's a PyObject* pointing to the
    singleton (we hope) _Py_NoneStruct with ob_type PyNone_Type
    """
    _typename = 'PyObject'

    def proxyval(self, visited):
        return None


class PyFrameObjectPtr(PyObjectPtr):
    _typename = 'PyFrameObject'

    def __init__(self, gdbval, cast_to):
        PyObjectPtr.__init__(self, gdbval, cast_to)

        if not self.is_optimized_out():
            self.co = PyCodeObjectPtr.from_pyobject_ptr(self.field('f_code'))
            self.co_name = self.co.pyop_field('co_name')
            self.co_filename = self.co.pyop_field('co_filename')
            self.f_lineno = int_from_int(self.field('f_lineno'))
            self.f_lasti = int_from_int(self.field('f_lasti'))
            self.co_nlocals = int_from_int(self.co.field('co_nlocals'))
            self.co_varnames = PyTupleObjectPtr.from_pyobject_ptr(
                self.co.field('co_varnames'))
            self.f_back = self.field('f_back')

    def iter_locals(self):
        """
        Yield a sequence of (name,value) pairs of PyObjectPtr instances, for
        the local variables of this frame
        """
        if self.is_optimized_out():
            return

        f_localsplus = self.field('f_localsplus')
        for i in safe_range(self.co_nlocals):
            pyop_value = PyObjectPtr.from_pyobject_ptr(f_localsplus[i])
            if not pyop_value.is_null():
                pyop_name = PyObjectPtr.from_pyobject_ptr(self.co_varnames[i])
                yield (pyop_name, pyop_value)

    def iter_globals(self):
        """
        Yield a sequence of (name,value) pairs of PyObjectPtr instances, for
        the global variables of this frame
        """
        if self.is_optimized_out():
            return

        pyop_globals = self.pyop_field('f_globals')
        return pyop_globals.iteritems()

    def iter_builtins(self):
        """
        Yield a sequence of (name,value) pairs of PyObjectPtr instances, for
        the builtin variables
        """
        if self.is_optimized_out():
            return

        pyop_builtins = self.pyop_field('f_builtins')
        return pyop_builtins.iteritems()

    def get_var_by_name(self, name):
        """
        Look for the named local variable, returning a (PyObjectPtr, scope) pair
        where scope is a string 'local', 'global', 'builtin'

        If not found, return (None, None)
        """
        for pyop_name, pyop_value in self.iter_locals():
            if name == pyop_name.proxyval(set()):
                return pyop_value, 'local'
        for pyop_name, pyop_value in self.iter_globals():
            if name == pyop_name.proxyval(set()):
                return pyop_value, 'global'
        for pyop_name, pyop_value in self.iter_builtins():
            if name == pyop_name.proxyval(set()):
                return pyop_value, 'builtin'
        return None, None

    def filename(self):
        """Get the path of the current Python source file, as a string"""
        if self.is_optimized_out():
            return '(frame information optimized out)'
        return self.co_filename.proxyval(set())

    def current_line_num(self):
        """Get current line number as an integer (1-based)

        Translated from PyFrame_GetLineNumber and PyCode_Addr2Line

        See Objects/lnotab_notes.txt
        """
        if self.is_optimized_out():
            return None
        f_trace = self.field('f_trace')
        if long(f_trace) != 0:
            # we have a non-NULL f_trace:
            return self.f_lineno
        else:
            # try:
            return self.co.addr2line(self.f_lasti)
            # except ValueError:
            #    return self.f_lineno

    def current_line(self):
        """Get the text of the current source line as a string, with a trailing
        newline character"""
        if self.is_optimized_out():
            return '(frame information optimized out)'
        with open(self.filename(), 'r') as f:
            all_lines = f.readlines()
            # Convert from 1-based current_line_num to 0-based list offset:
            return all_lines[self.current_line_num()-1]

    def write_repr(self, out, visited):
        if self.is_optimized_out():
            out.write('(frame information optimized out)')
            return
        out.write('Frame 0x%x, for file %s, line %i, in %s ('
                  % (self.as_address(),
                     self.co_filename,
                     self.current_line_num(),
                     self.co_name))
        first = True
        for pyop_name, pyop_value in self.iter_locals():
            if not first:
                out.write(', ')
            first = False

            out.write(pyop_name.proxyval(visited))
            out.write('=')
            pyop_value.write_repr(out, visited)

        out.write(')')


class PySetObjectPtr(PyObjectPtr):
    _typename = 'PySetObject'

    def proxyval(self, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            return ProxyAlreadyVisited('%s(...)' % self.safe_tp_name())
        visited.add(self.as_address())

        members = []
        table = self.field('table')
        for i in safe_range(self.field('mask')+1):
            setentry = table[i]
            key = setentry['key']
            if key != 0:
                key_proxy = PyObjectPtr.from_pyobject_ptr(
                    key).proxyval(visited)
                if key_proxy != '<dummy key>':
                    members.append(key_proxy)
        if self.safe_tp_name() == 'frozenset':
            return frozenset(members)
        else:
            return set(members)

    def write_repr(self, out, visited):
        out.write(self.safe_tp_name())

        # Guard against infinite loops:
        if self.as_address() in visited:
            out.write('(...)')
            return
        visited.add(self.as_address())

        out.write('([')
        first = True
        table = self.field('table')
        for i in safe_range(self.field('mask')+1):
            setentry = table[i]
            key = setentry['key']
            if key != 0:
                pyop_key = PyObjectPtr.from_pyobject_ptr(key)
                key_proxy = pyop_key.proxyval(visited)  # FIXME!
                if key_proxy != '<dummy key>':
                    if not first:
                        out.write(', ')
                    first = False
                    pyop_key.write_repr(out, visited)
        out.write('])')


class PyStringObjectPtr(PyObjectPtr):
    _typename = 'PyStringObject'

    def __str__(self):
        field_ob_size = self.field('ob_size')
        field_ob_sval = self.field('ob_sval')
        char_ptr = field_ob_sval.address.cast(_type_unsigned_char_ptr)
        return ''.join([chr(char_ptr[i]) for i in safe_range(field_ob_size)])

    def proxyval(self, visited):
        return str(self)


class PyTupleObjectPtr(PyObjectPtr):
    _typename = 'PyTupleObject'

    def __getitem__(self, i):
        # Get the gdb.Value for the (PyObject*) with the given index:
        field_ob_item = self.field('ob_item')
        return field_ob_item[i]

    def proxyval(self, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            return ProxyAlreadyVisited('(...)')
        visited.add(self.as_address())

        result = tuple([PyObjectPtr.from_pyobject_ptr(
            self[i]).proxyval(visited)
            for i in safe_range(int_from_int(self.field('ob_size')))])
        return result

    def write_repr(self, out, visited):
        # Guard against infinite loops:
        if self.as_address() in visited:
            out.write('(...)')
            return
        visited.add(self.as_address())

        out.write('(')
        for i in safe_range(int_from_int(self.field('ob_size'))):
            if i > 0:
                out.write(', ')
            element = PyObjectPtr.from_pyobject_ptr(self[i])
            element.write_repr(out, visited)
        if self.field('ob_size') == 1:
            out.write(',)')
        else:
            out.write(')')


class PyTypeObjectPtr(PyObjectPtr):
    _typename = 'PyTypeObject'


class PyUnicodeObjectPtr(PyObjectPtr):
    _typename = 'PyUnicodeObject'

    def proxyval(self, visited):
        # From unicodeobject.h:
        #     Py_ssize_t length;  /* Length of raw Unicode data in buffer */
        #     Py_UNICODE *str;    /* Raw Unicode buffer */
        field_length = long(self.field('length'))
        field_str = self.field('str')

        # Gather a list of ints from the Py_UNICODE array; these are either
        # UCS-2 or UCS-4 code points:
        Py_UNICODEs = [int(field_str[i]) for i in safe_range(field_length)]

        # Convert the int code points to unicode characters, and generate a
        # local unicode instance:
        result = u''.join([unichr(ucs) for ucs in Py_UNICODEs])
        return result


def int_from_int(gdbval):
    return int(str(gdbval))


class Frame(object):
    """
    Wrapper for gdb.Frame, adding various methods
    """

    def __init__(self, gdbframe):
        self._gdbframe = gdbframe

    def older(self):
        older = self._gdbframe.older()
        if older:
            return Frame(older)
        else:
            return None

    def newer(self):
        newer = self._gdbframe.newer()
        if newer:
            return Frame(newer)
        else:
            return None

    def select(self):
        self._gdbframe.select()

    def get_index(self):
        """Calculate index of frame, starting at 0 for the newest frame within
        this thread"""
        index = 0
        # Go down until you reach the newest frame:
        iter_frame = self
        while iter_frame.newer():
            index += 1
            iter_frame = iter_frame.newer()
        return index

    def is_evalframeex(self):
        if self._gdbframe.function():
            if self._gdbframe.function().name == 'PyEval_EvalFrameEx':
                """
                I believe we also need to filter on the inline
                struct frame_id.inline_depth, only regarding frames with
                an inline depth of 0 as actually being this function

                So we reject those with type gdb.INLINE_FRAME
                """
                if self._gdbframe.type() == gdb.NORMAL_FRAME:
                    # We have a PyEval_EvalFrameEx frame:
                    return True

        return False

    def get_pyop(self):
        try:
            f = self._gdbframe.read_var('f')
            return PyFrameObjectPtr.from_pyobject_ptr(f)
        except ValueError:
            return None

    @classmethod
    def get_selected_frame(cls):
        _gdbframe = gdb.selected_frame()
        if _gdbframe:
            return Frame(_gdbframe)
        return None

    @classmethod
    def get_selected_python_frame(cls):
        """Try to obtain the Frame for the python code in the selected frame,
        or None"""
        frame = cls.get_selected_frame()

        while frame:
            if frame.is_evalframeex():
                return frame
            frame = frame.older()

        # Not found:
        return None

    def print_summary(self):
        if self.is_evalframeex():
            pyop = self.get_pyop()
            if pyop:
                sys.stdout.write('#%i %s\n' % (
                    self.get_index(), pyop.get_truncated_repr(MAX_OUTPUT_LEN)))
                sys.stdout.write(pyop.current_line())
            else:
                sys.stdout.write(
                    '#%i (unable to read python frame information)\n'
                    % self.get_index())
        else:
            sys.stdout.write('#%i\n' % self.get_index())


def move_in_stack(move_up, silently=False):
    """Move up or down the stack (for the py-up/py-down command)"""
    frame = Frame.get_selected_python_frame()
    while frame:
        if move_up:
            iter_frame = frame.older()
        else:
            iter_frame = frame.newer()

        if not iter_frame:
            break

        if iter_frame.is_evalframeex():
            # Result:
            iter_frame.select()
            if not silently:
                iter_frame.print_summary()
            return True

        frame = iter_frame

    if not silently:
        if move_up:
            print 'Unable to find an older python frame'
        else:
            print 'Unable to find a newer python frame'

    return False



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# --- BEGIN_HEADER ---
#
# pygdb.console.extension - python gdb console extension functions
# Copyright (C) 2019-2020  The pygdb Project lead by Brian Vinter
#
# This file is part of pygdb.
#
# pygdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pygdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# -- END_HEADER ---
#

"""GDB console python extension functions"""

import sys
import gdb
from pygdb.console.core import Frame, move_in_stack, \
    PyDictObjectPtr, PyInstanceObjectPtr, HeapTypeObjectPtr

__breakpoint_identifier = '_pygdb_breakpoint_mark'
__breakpoint_func = '_pygdb.breakpoint_mark()'


def term_color(*v):
    """
    set_term_color(text-type, text-color, background-color)
    text-type: 1: bold
               2: faded
               3: italic
               4: underlined
    text-color: 30: black
                31: red
                32: green
                33: yellow
                34: blue
                35: magneta
                36: cyan
    back-ground-color 40: black
                      41: red
                      42: green
                      43: yellow
                      44: blue
                      45: magneta
                      46: cyan
    set_term_color(0): #reset
    """

    return '\x1B['+';'.join(map(str, v))+'m'


def inject_pyframe(cmd, silently=False):
    """Inject python code *cmd* into selected frame"""

    result = gdb.execute("call PyGILState_Ensure()", to_string=True)
    gstate = result.split('=')[1].strip()
    if not silently:
        print "PyGILState_Ensure: %s" % gstate
    gdb.execute("call PyRun_SimpleString(\"%s\")" % cmd)
    gdb.execute("call PyGILState_Release(%s)" % gstate)


def get_pyframe_f_back(silently=False):
    """Get id of parent pyframe"""

    pyop_frame = get_selected_pyop()
    if not pyop_frame:
        return None

    return pyop_frame.f_back


def get_selected_pyop(silently=False):
    """Returns pyop for selected frame"""

    frame = Frame.get_selected_python_frame()
    if not frame:
        if not silently:
            print 'Unable to locate python frame'
        return None

    pyop = frame.get_pyop()
    if not pyop:
        if not silently:
            print 'Unable to read information on python frame'
        return None

    return pyop


def get_pyobject_value(name):
    """Returns pyobject value for *name*,
    supports nested pyobject values are supported"""

    name_arr = name.split('.')
    name_arr_len = len(name_arr)
    first_name = name_arr[0]

    pyop_frame = get_selected_pyop()
    if not pyop_frame:
        return (None, None)

    (pyop_var, scope) = pyop_frame.get_var_by_name(first_name)
    if not pyop_var:
        return (scope, None)

    cur_value = pyop_var
    for i in xrange(1, name_arr_len):
        next_name = name_arr[i]
        pyop_var_dict = None
        if isinstance(cur_value, PyDictObjectPtr):
            pyop_var_dict = pyop_var
        elif isinstance(cur_value, PyInstanceObjectPtr):
            pyop_var_dict = pyop_var.pyop_field('in_dict')
        elif isinstance(cur_value, HeapTypeObjectPtr):
            pyop_var_dict = cur_value.get_attr_dict()
        else:
            return (scope, None)
        cur_value = None
        if pyop_var_dict:
            # NOTE: pyop_var_dict does not support __getitem__
            for key, value in pyop_var_dict.iteritems():
                if str(key) == next_name:
                    cur_value = value
        if cur_value is None:
            break

    return (scope, cur_value)


def breakpoint_frame(silently=False):
    """Move to python frame with active breakpoint"""

    org_frame = gdb.selected_frame()
    gdb.newest_frame().select()
    pyop = get_selected_pyop()
    if not pyop:
        org_frame.select()
        return False
    line = pyop.current_line().strip()
    if line != __breakpoint_func:
        if not silently:
            print "No breakpoint found for current thread"
        org_frame.select()
        return False
    elif not silently:
        frame = Frame.get_selected_python_frame()
        frame.print_summary()

    return True


def breakpoint_caller_frame(silently=False):
    """Move to the frame that called the breakpoint frame"""
    status = breakpoint_frame(silently=silently)
    if status:
        more_frames = move_in_stack(move_up=True, silently=silently)
        if not more_frames:
            status = False

    return status


def list_pyframe(start=None, end=None):
    """List current python frame with active code line highlighted"""
    show_lines = 30
    pyop = get_selected_pyop()
    if not pyop:
        return

    filename = pyop.filename()
    lineno = pyop.current_line_num()

    if start is None:
        start = lineno - (show_lines/2)
        end = lineno + (show_lines/2)
    if end is None:
        end = start + show_lines
    if start < 1:
        start = 1

    thread = gdb.execute('thread', to_string=True)
    sys.stdout.write("%s%s:%s\n" % (thread, filename, lineno))
    with open(filename, 'r') as fh:
        all_lines = fh.readlines()
        # start and end are 1-based, all_lines is 0-based;
        for i, line in enumerate(all_lines[start-1:end]):

            display_lineno = int(i+start)
            # Highlight current line:
            # print color + sys.stdout.write(line[:-1])
            # is a hack to color the full line
            if display_lineno == lineno:
                color = term_color(0, 34, 47)
            else:
                color = term_color(0)
            print color
            sys.stdout.write("%d\t%s" % (display_lineno, line[:-1]))
    color = term_color(0)
    print color
    print ""
    fh.close()


def attach(pid):
    """Delete old breakpoints, add python breakpoint,
    attach process and signal that GDB console is attached"""

    # Delete old breakpoints
    gdb.execute('delete breakpoints')

    # Attach process
    gdb.execute('attach %d' % pid)

    # Add python breakpoint to breakpoints
    gdb.execute('break %s' % __breakpoint_identifier)

    # Signal that GDB console is connected (handled by pygdb.breakpoint)
    gdb.execute('signal SIGCONT')

    # Show breakpoint
    breakpoint_list()


def list_threads():
    """List all threads for attached process"""
    # NOTE: threads[-1] is main thread
    threads = gdb.selected_inferior().threads()
    for thread in threads:
        print thread


def switch_thread(threadid, silently=False):
    """Change active thread to *threadid*"""
    cmd = "thread find Thread 0x%.x" % threadid
    thread_info = gdb.execute(cmd, to_string=True)
    thread_info_arr = thread_info.split(" ")
    if thread_info_arr[0] == 'No':
        if not silently:
            print "No Thread found with thread id: 0x%.x" % threadid
        return False
    else:
        cmd = "thread %s" % thread_info_arr[1]
        thread_switch = gdb.execute(cmd, to_string=True)
        if not silently:
            print thread_switch
        return True


def breakpoint_list(threadid=None, silently=False):
    """Display python code for breakpoint frame"""

    status = breakpoint_caller_frame(silently=True)
    if status:
        list_pyframe()
    elif not silently:
        if threadid:
            msg = "Unable to find breakpoint for thread: 0x%.x" % threadid
        else:
            msg = "Unable to find breakpoint for current thread"
        print msg


def breakpoint_continue():
    """Continue until next breakpoint and display it"""
    gdb.execute('continue')
    breakpoint_list()


def inspect_pyframe(show_globals=True, show_locals=True):
    """Display details about the current python frame"""

    pyframe_index = 0
    pyframe_f_back = get_pyframe_f_back()

    arguments = {'pyframe_index': pyframe_index,
                 'pyframe_f_back': pyframe_f_back,
                 'show_globals': show_globals,
                 'show_locals': show_locals}

    print "===================================================================="
    print "Output is written to gdb_logger if found otherwise to process stdout"
    print "===================================================================="

    cmd = \
        "from __future__ import print_function as __GDB_DEBUG_print_function; \
        import inspect as __GDB_DEBUG_inspect; \
        __GDB_DEBUG_logger = \
            print if not 'pygdb' in globals() \
            else pygdb.breakpoint.gdb_logger_debug; \
        __GDB_DEBUG_stack = __GDB_DEBUG_inspect.stack(); \
        __GDB_DEBUG_stack_len = len(__GDB_DEBUG_stack); \
        __GDB_DEBUG_frame = __GDB_DEBUG_stack[%(pyframe_index)d][0]; \
        __GDB_DEBUG_frameidx_func = lambda frame: \
            0 if not frame or id(frame.f_back) == %(pyframe_f_back)d \
            else 1 + __GDB_DEBUG_frameidx_func(frame.f_back); \
        __GDB_DEBUG_frameidx = __GDB_DEBUG_frameidx_func(__GDB_DEBUG_frame); \
        __GDB_DEBUG_frameidx = __GDB_DEBUG_frameidx \
            if __GDB_DEBUG_frameidx < __GDB_DEBUG_stack_len \
            else 0; \
        __GDB_DEBUG_frame, \
        __GDB_DEBUG_filename, \
        __GDB_DEBUG_line_number, \
        __GDB_DEBUG_function_name, \
        __GDB_DEBUG_lines, \
        __GDB_DEBUG_index = __GDB_DEBUG_stack[__GDB_DEBUG_frameidx]; \
        __GDB_DEBUG_logger(chr(10) \
            + '========================== py_inspect_frame ==========================' + chr(10) \
            + 'frame No.: ' + str(__GDB_DEBUG_frameidx) + chr(10) \
            + 'frame: ' + str(__GDB_DEBUG_frame) + chr(10) \
            + 'frame.f_back: ' + str(__GDB_DEBUG_frame.f_back) + chr(10) \
            + 'filename: ' + str(__GDB_DEBUG_filename) + chr(10) \
            + 'line_number: ' + str(__GDB_DEBUG_line_number) + chr(10) \
            + 'function_name: ' + str(__GDB_DEBUG_function_name) + chr(10) \
            + 'lines: ' + str(__GDB_DEBUG_lines) + chr(10) \
            + 'index: ' + str(__GDB_DEBUG_index) + chr(10) \
            + '----------------------------------------------------------------------' + chr(10) \
            + 'f_globals: ' + chr(10) \
            + '----------------------------------------------------------------------' + chr(10) \
            + chr(10).join(' = '.join((str(k),str(v))) for k,v in __GDB_DEBUG_frame.f_globals.items()\
                                if %(show_globals)s and not k.startswith('__GDB_DEBUG')) + chr(10) \
            + '----------------------------------------------------------------------' + chr(10) \
            + 'f_locals: ' + chr(10) \
            + '----------------------------------------------------------------------' + chr(10) \
            + chr(10).join(' = '.join((str(k),str(v))) for k,v in __GDB_DEBUG_frame.f_locals.items() \
                                if %(show_locals)s and not k.startswith('__GDB_DEBUG')) + chr(10) \
            + '======================== End py_inspect_frame ========================' + chr(10)); \
            [globals().pop(key) \
                for key in globals().keys() \
                    if key.startswith('__GDB_DEBUG_')]" % arguments

    inject_pyframe(cmd)


def set_pyframe_local(key, value):
    """Set local variable *key* to *value* in current python frame"""
    arguments = {'pyframe_index': 0,
                 'pyframe_f_back': get_pyframe_f_back(),
                 'key': key,
                 'value': value}
    cmd = \
        "from __future__ import print_function as __GDB_DEBUG_print_function; \
        import inspect as __GDB_DEBUG_inspect; \
        import ctypes as __GDB_DEBUG_ctypes; \
        __GDB_DEBUG_logger = \
            print if not 'pygdb' in globals() \
            else pygdb.breakpoint.gdb_logger_debug; \
        __GDB_DEBUG_stack = __GDB_DEBUG_inspect.stack(); \
        __GDB_DEBUG_stack_len = len(__GDB_DEBUG_stack); \
        __GDB_DEBUG_frame = __GDB_DEBUG_stack[%(pyframe_index)d][0]; \
        __GDB_DEBUG_frameidx_func = lambda frame: \
            0 if not frame or id(frame.f_back) == %(pyframe_f_back)d \
            else 1 + __GDB_DEBUG_frameidx_func(frame.f_back); \
        __GDB_DEBUG_frameidx = __GDB_DEBUG_frameidx_func(__GDB_DEBUG_frame); \
        __GDB_DEBUG_frameidx = __GDB_DEBUG_frameidx \
            if __GDB_DEBUG_frameidx < __GDB_DEBUG_stack_len \
            else 0; \
        __GDB_DEBUG_frame = __GDB_DEBUG_stack[__GDB_DEBUG_frameidx][0]; \
        __GDB_DEBUG_frame.f_locals['%(key)s'] = %(value)s; \
        __GDB_DEBUG_ctypes.pythonapi.PyFrame_LocalsToFast(\
                     __GDB_DEBUG_ctypes.py_object(__GDB_DEBUG_frame),\
                                              __GDB_DEBUG_ctypes.c_int(0)); \
        [globals().pop(key) \
            for key in globals().keys() \
                if key.startswith('__GDB_DEBUG_')]" % arguments

    inject_pyframe(cmd)


def pystep(skip_breakpoint_mark=True,
           file_prefix=None,
           silently=False,
           scheduler_locking=False):
    """Continue until control reaches a different python source line"""

    if scheduler_locking:
        status = gdb.execute('set scheduler-locking step', to_string=True)
    else:
        status = gdb.execute('set scheduler-locking off', to_string=True)
    if not silently:
        print status

    step_count = 0
    max_step_count = 10000
    pyop = get_selected_pyop(silently=silently)
    if not pyop:
        return
    start_filepath = filepath = pyop.filename()
    start_lineno = lineno = pyop.current_line_num()
    line = pyop.current_line().strip()
    while (start_filepath == filepath and start_lineno == lineno) \
            or (skip_breakpoint_mark and line == __breakpoint_func) \
            or (file_prefix and not filepath.startswith(file_prefix)):
        step_msg = gdb.execute('step', to_string=True)
        gdb.newest_frame().select()
        pyop = get_selected_pyop(silently=silently)
        if not pyop:
            break
        filepath = pyop.filename()
        lineno = pyop.current_line_num()
        line = pyop.current_line().strip()
        step_count += 1
        if step_count > max_step_count:
            break

    if not silently:
        print "#steps: %s" % step_count
        list_pyframe()

    if step_count > max_step_count:
        sys.stdout.write(term_color(0, 31))
        print "WARNING: Stopped after max steps: %s" % max_step_count
        sys.stdout.write(term_color(0))


def pynext(skip_breakpoint_mark=True,
           file_prefix=None,
           scheduler_locking=False,
           silently=False):
    """Continue until control reaches a different python source line in current frame"""

    if scheduler_locking:
        status = gdb.execute('set scheduler-locking on', to_string=True)
    else:
        status = gdb.execute('set scheduler-locking off', to_string=True)
    if not silently:
        print status

    next_count = 0
    max_next_count = 10000
    pyop = get_selected_pyop(silently=silently)
    if not pyop:
        return
    start_filepath = filepath = pyop.filename()
    start_lineno = lineno = pyop.current_line_num()
    line = pyop.current_line().strip()
    while (start_filepath == filepath and start_lineno == lineno) \
            or (skip_breakpoint_mark and line == __breakpoint_func) \
            or (file_prefix and not filepath.startswith(file_prefix)):
        next_msg = gdb.execute('next', to_string=True)
        gdb.newest_frame().select()
        pyop = get_selected_pyop(silently=silently)
        if not pyop:
            break
        filepath = pyop.filename()
        lineno = pyop.current_line_num()
        line = pyop.current_line().strip()
        next_count += 1

        if next_count > max_next_count:
            break

    if not silently:
        print "#next: %s" % next_count
        list_pyframe()

    if next_count > max_next_count:
        sys.stdout.write(term_color(0, 31))
        print "WARNING: Stopped after max next's: %s" % max_next_count
        sys.stdout.write(term_color(0))



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# --- BEGIN_HEADER ---
#
# __init__ - gdb python extensions
# Copyright (C) 2019-2020  The pygdb Project lead by Brian Vinter
#
# This file is part of pygdb.
#
# pygdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pygdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# -- END_HEADER ---
	#

"""This package provide python extensions for gdb. 
This file is needed to tell python that this dir is a package
so that other modules can call, say, import pygdb.breakpoint.
Please refer to http://www.network-theory.co.uk/docs/pytut/tut_51.html
for details
"""

__dummy = True

# above line is only to make python tidy behave and not
# move module doc string inside header

# All sub modules to load in case of 'from X import *'

__all__ = [
	'breakpoint', 
	'console',
	]

# Collect all package information here for easy use from scripts and helpers

package_name = 'GDB Python extension'
short_name = 'pygdb'

# IMPORTANT: Please keep version in sync with doc-src/README.t2t

version_tuple = (0, 1, 0)
version_suffix = 'post2'
version_string = '.'.join([str(i) for i in version_tuple]) + version_suffix
package_version = '%s %s' % (package_name, version_string)
project_team = 'The pygdb Project lead by Brian Vinter'
project_email = 'pygdb@erda.dk'
maintainer_team = 'The pygdb maintainers'
maintainer_email = 'pygdb@erda.dk'
project_url = 'https://github.com/ucphhpc/pygdb'
download_url = 'https://github.com/ucphhpc/pygdb/releases'
license_name = 'GNU GPL v2'
short_desc = \
    'Python extension for the GNU project debugger (GDB)'
long_desc = \
    """Python extension for the GNU project debugger (GDB):
https://www.gnu.org/software/gdb/

Documentation: https://github.com/ucphhpc/pygdb
"""
project_class = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
    'Operating System :: OS Independent',
    'Programming Language :: Python :: 2.7',
    'Topic :: Software Development :: Debuggers',
    ]
project_keywords = [
    'Python',
    'Python C extensions',
    'Debugger',
    ]

# Requirements

full_requires = []
versioned_requires = []
project_requires = []

# Optional packages required for additional functionality (for extras_require)

project_extras = {}
package_provides = short_name
project_platforms = ['All']



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# --- BEGIN_HEADER ---
#
# breakpoint - Python GDB breakpoint
# Copyright (C) 2019-2020  The pygdb Project lead by Brian Vinter
#
# This file is part of pygdb.
#
# pygdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pygdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# -- END_HEADER ---
#

"""Python GNU debugger (GDB) breakpoint
https://devguide.python.org/gdb/
https://wiki.python.org/moin/DebuggingWithGdb
https://sourceware.org/gdb/onlinedocs/gdb/Python-API.html

USAGE:
import pygdb.breakpoint
pygdb.breakpoint.enable()
pygdb.breakpoint.set()

This will block the executing thread (at the pygdb.breakpoint.set() call)
until 'pygdb.console.extension.console_connected' is called
from the GDB console.
"""

import os
import sys
import logging
import signal
import time
import threading
import _pygdb

enabled = False
gdb_logger = None
console_connected = None
breakpoint_lock = threading.Lock()


def handle_sigcont(signalNumber, frame):
    """ NOTE: We do not forward SIGCONT because: 
    You can set a handler, but SIGCONT always makes the process continue regardless:
    https://www.gnu.org/software/libc/manual/html_node/Job-Control-Signals.html 
    """
    global console_connected
    breakpoint_lock.acquire()
    if not console_connected:
        console_connected = True
        log("GDB console attached")
    breakpoint_lock.release()


def enable(logger=None):
    """Enable breakpoint"""
    global enabled
    global console_connected
    global breakpoint_lock

    breakpoint_lock.acquire()
    if not enabled:
        enabled = True
        console_connected = False
        set_logger(logger)
        signal.signal(signal.SIGCONT, handle_sigcont)
        log("enabled")
    breakpoint_lock.release()


def log(msg):
    """log info messages"""
    if not enabled:
        return

    pid = os.getpid()
    tid = threading.current_thread().ident
    log_msg = "pygdb: (PID: %d, TID: 0x%0.x): %s" \
        % (pid, tid, msg)

    if isinstance(gdb_logger, logging.Logger):
        gdb_logger.info(log_msg)
    else:
        sys.stderr.write("pygdb: %s\n" % log_msg)


def set_logger(logger):
    """set logger to used by the log function"""
    if not enabled:
        return False

    global gdb_logger

    result = False
    if isinstance(logger, logging.Logger):
        gdb_logger = logger
        result = True

    return result


def set(logger=None):
    """Used to set breakpoint, busy-wait until gdb console is connected"""
    if not enabled:
        return

    global console_connected

    # Wait for gdb console
    breakpoint_lock.acquire()
    if logger is not None:
        set_logger(logger)
    while not console_connected:
        log("breakpoint.set: waiting for gdb console")
        breakpoint_lock.release()
        time.sleep(1)
        breakpoint_lock.acquire()
    breakpoint_lock.release()
    _pygdb.breakpoint_mark()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# --- BEGIN_HEADER ---
#
# setup.py - Setup for Python GDB debugger
# Copyright (C) 2019-2020  The pygdb Project lead by Brian Vinter
#
# This file is part of pygdb.
#
# pygdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pygdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# -- END_HEADER ---
#
from setuptools import setup, Extension

from pygdb import version_string, short_name, project_team, \
    project_email, short_desc, long_desc, project_url, download_url, \
    license_name, project_class, project_keywords, versioned_requires, \
    project_requires, project_extras, project_platforms, maintainer_team, \
    maintainer_email

setup(
    name=short_name,
    version=version_string,
    description=short_desc,
    long_description=long_desc,
    author=project_team,
    author_email=project_email,
    maintainer=maintainer_team,
    maintainer_email=maintainer_email,
    url=project_url,
    download_url=download_url,
    license=license_name,
    classifiers=project_class,
    keywords=project_keywords,
    platforms=project_platforms,
    install_requires=versioned_requires,
    requires=project_requires,
    extras_require=project_extras,
    packages=['pygdb', 'pygdb.console'],
    package_dir={'pygdb': 'pygdb',
                 'pygdb.console': 'pygdb/console'},
    ext_modules = [
        Extension('_pygdb', ['pygdb/_breakpoint.c'],
          define_macros=[('NDEBUG', '0')],),
        ]
    )


#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# --- BEGIN_HEADER ---
#
# __init__ - gdb python console extensions
# Copyright (C) 2019-2020  The pygdb Project lead by Brian Vinter
#
# This file is part of pygdb.
#
# pygdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pygdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# -- END_HEADER ---
#

"""This package provide python extensions for gdb console. 
This file is needed to tell python that this dir is a package
so that other modules can call, say, import pygdb.breakpoint.
Please refer to http://www.network-theory.co.uk/docs/pytut/tut_51.html
for details
"""

__dummy = True

# above line is only to make python tidy behave and not
# move module doc string inside header



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# --- BEGIN_HEADER ---
#
# pygdb.console.commands - python gdb console commands
# Copyright (C) 2019-2020  The pygdb Project lead by Brian Vinter
#
# This file is part of pygdb.
#
# pygdb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pygdb is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# -- END_HEADER ---
#

"""Python gdb console commands"""

import re
import gdb
from pygdb.console.core import Frame, move_in_stack, \
    PyObjectPtr, PyDictObjectPtr, PyInstanceObjectPtr, HeapTypeObjectPtr, \
    MAX_OUTPUT_LEN

from pygdb.console.extensions import attach, \
    breakpoint_continue, breakpoint_list, \
    get_pyobject_value, inject_pyframe, inspect_pyframe, \
    list_pyframe, pystep, pynext, set_pyframe_local, switch_thread

commands = None


def help():
    """Display help for all commands"""
    global commands
    if not commands:
        register()
    for cmd in commands:
        print "-------------------------------------------------------------------------------"
        cmd.help()
    print "-------------------------------------------------------------------------------"


def register():
    """Register new commands to the GDB console"""
    global commands
    commands = [
        cmd_py_info_procs('py-info-procs'),
        cmd_py_info_threads('py-info-threads'),
        cmd_py_attach('py-attach'),
        cmd_py_detach('py-detach'),
        cmd_py_thread('py-thread'),
        cmd_py_breakpoint('py-breakpoint'),
        cmd_py_step('py-step'),
        cmd_py_next('py-next'),
        cmd_py_continue('py-continue'),
        cmd_py_list('py-list'),
        cmd_py_backtrace('py-backtrace'),
        cmd_py_inspect_frame('py-inspect-frame'),
        cmd_py_builtins('py-builtins'),
        cmd_py_globals('py-globals'),
        cmd_py_locals('py-locals'),
        cmd_py_print('py-print'),
        cmd_py_set_local('py-set-local'),
        cmd_py_up('py-up'),
        cmd_py_down('py-down'),
        cmd_py_inject("py-inject")
    ]


class cmd_py_info_procs(gdb.Command):
    """Display list of attached processes"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return
        gdb.execute('info inferior')


class cmd_py_info_threads(gdb.Command):
    """Display list of attached process threads"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return
        gdb.execute('info threads')


class cmd_py_attach(gdb.Command):
    """Attach python process (PID) to gdb console"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s PID" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return
        try:
            pid = int(args)
        except:
            self.help()
            return

        # Reset GDB environment including old breakpoints

        try:
            gdb.execute('delete')
        except Exception:
            pass

        # Attach pid

        attach(pid)

        # Re-register pygdb commands

        register()


class cmd_py_detach(gdb.Command):
    """Detach python process (GDB-PROCESS-NUM) from gdb console"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s GDB-PROCESS-NUM" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return
        try:
            num = int(args)
        except:
            self.help()
            return
        gdb.execute("detach inferior %d" % num)


class cmd_py_thread(gdb.Command):
    """Switch to python thread id (TID)"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s TID" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return
        try:
            threadid = int(args, 16)
        except:
            self.help()
            return
        switch_thread(threadid)


class cmd_py_breakpoint(gdb.Command):
    """Find and display code for python breakpoint
    Switch to optional python thread id (TID) if provided"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s [TID]" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return
        threadid = None
        try:
            if args:
                threadid = int(args, 16)
        except:
            self.help()
            return
        breakpoint_list(threadid=threadid)


class cmd_py_step(gdb.Command):
    """Continue until control reaches a different python source line"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        file_prefix = None
        if args:
            file_prefix = args
        pystep(file_prefix=file_prefix)


class cmd_py_next(gdb.Command):
    """Continue until control reaches a different python source line in current frame"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        file_prefix = None
        if args:
            file_prefix = args
        pynext(file_prefix=file_prefix)


class cmd_py_continue(gdb.Command):
    """Continue until next python breakpoint and display code"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return
        breakpoint_continue()


class cmd_py_list(gdb.Command):
    """Display code for current python frame

    Use argument:
       START
    to list at a different line number within the python source.

    Use argument:
       START, END
    to list a specific range of lines within the python source.
    """

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_FILES,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s [start] [end]" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        start = None
        end = None

        m = re.match(r'\s*(\d+)\s*', args)
        if m:
            start = int(m.group(0))

        m = re.match(r'\s*(\d+)\s*,\s*(\d+)\s*', args)
        if m:
            start, end = map(int, m.groups())

        list_pyframe(start=start, end=end)


class cmd_py_backtrace(gdb.Command):
    """Display current python frame and all the frames within its call stack (if any)"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_STACK,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        frame = Frame.get_selected_python_frame()
        while frame:
            if frame.is_evalframeex():
                frame.print_summary()
            frame = frame.older()


class cmd_py_inspect_frame(gdb.Command):
    """Display information about the current python frame"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s [start] [end]" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        inspect_pyframe()


class cmd_py_builtins(gdb.Command):
    """Display builtin varialbes for current python frame"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        frame = Frame.get_selected_python_frame()
        if not frame:
            print "Unable to locate python frame"
            return

        pyop_frame = frame.get_pyop()
        if not pyop_frame:
            print "Unable to read information on python frame"
            return

        for pyop_name, pyop_value in pyop_frame.iter_builtins():
            print('%s = %s'
                  % (pyop_name.proxyval(set()),
                      pyop_value.get_truncated_repr(MAX_OUTPUT_LEN)))


class cmd_py_globals(gdb.Command):
    """Display global variables for current python frame"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        frame = Frame.get_selected_python_frame()
        if not frame:
            print "Unable to locate python frame"
            return

        pyop_frame = frame.get_pyop()
        if not pyop_frame:
            print "Unable to read information on python frame"
            return

        for pyop_name, pyop_value in pyop_frame.iter_globals():
            print('%s = %s'
                  % (pyop_name.proxyval(set()),
                      pyop_value.get_truncated_repr(MAX_OUTPUT_LEN)))


class cmd_py_locals(gdb.Command):
    """Display local variables for current python frame"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        frame = Frame.get_selected_python_frame()
        if not frame:
            print "Unable to locate python frame"
            return

        pyop_frame = frame.get_pyop()
        if not pyop_frame:
            print "Unable to read information on python frame"
            return

        for pyop_name, pyop_value in pyop_frame.iter_locals():
            print('%s = %s'
                  % (pyop_name.proxyval(set()),
                      pyop_value.get_truncated_repr(MAX_OUTPUT_LEN)))


class cmd_py_print(gdb.Command):
    """Display python frame variable (if it exists)
    Nested dictionary and object values are displayed using dots

    Use argument:
        x
    to display value for variable x

    Use argument:
        x.y
    to display value y for dictionary/object variable x
    """

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        name = str(args)
        if name == "--help":
            self.help()
            return

        (scope, value) = get_pyobject_value(name)
        if value:
            value_dict = {}
            if isinstance(value, PyDictObjectPtr):
                value_dict = value
            elif isinstance(value, PyInstanceObjectPtr):
                value_dict = value.pyop_field('in_dict')
            elif isinstance(value, HeapTypeObjectPtr):
                value_dict = value.get_attr_dict()
            if value_dict:
                print "------------------------------------------------------------------------------"
                print "(%s) %s:" % (scope, name)
                print "------------------------------------------------------------------------------"
                for key, val in value_dict.iteritems():
                    if isinstance(key, PyObjectPtr):
                        key = key.get_truncated_repr(MAX_OUTPUT_LEN)
                    print "%s = %s" % (key, val.get_truncated_repr(MAX_OUTPUT_LEN))
                print "------------------------------------------------------------------------------"
            else:
                print "(%s) %s = %s" \
                    % (scope,
                       name,
                       value.get_truncated_repr(MAX_OUTPUT_LEN))
        else:
            print "%r not found" % name


class cmd_py_set_local(gdb.Command):
    """Set local variable in current python frame"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        key, value = args.split(" ")
        set_pyframe_local(key, value)


class cmd_py_up(gdb.Command):
    """Select and display code for the python stack frame that called this one (if any)"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_STACK,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        result = move_in_stack(move_up=True, silently=True)
        if result:
            list_pyframe()
        else:
            print "Unable to find an older python frame"


class cmd_py_down(gdb.Command):
    """Select and display code for the python stack frame called by this one (if any)"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_STACK,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        result = move_in_stack(move_up=False, silently=True)
        if result:
            list_pyframe()
        else:
            print "Unable to find an newer python frame"


class cmd_py_inject(gdb.Command):
    """Inject python code into current python frame"""

    name = None

    def __init__(self, name):
        self.name = name
        gdb.Command.__init__(self,
                             self.name,
                             gdb.COMMAND_DATA,
                             gdb.COMPLETE_COMMAND)

    def help(self):
        print "USAGE: %s" % self.name
        print ""
        doc_arr = self.__doc__.split('\n')
        for ent in doc_arr:
            print ent.strip()

    def invoke(self, args, from_tty):
        if str(args) == "--help":
            self.help()
            return

        cmd = str(args)
        inject_pyframe(cmd)
